#include <map>
#include <algorithm>
#include <stdexcept>

#include "mahjong.hpp"
#include "util.hpp"

using namespace std;

#define MAP_GET(m, k) ((m).find(k)->second)

namespace{
    int hand_size(const Hand& hand)
    {
        int size = 0;
        for(Pai p = PAI_MAN1; p != PAI_INVALID; p = pai_next_seq(p))
            size += MAP_GET(hand, p);
        return size;
    }

    typedef bool (*PickFunc)(const Hand& hand, Pai pai, Hand& nhand);
    bool pick_toitsu(const Hand& hand, Pai pai, Hand& nhand)
    {
        if(MAP_GET(hand, pai) < 2) return false;

        nhand = hand;
        nhand[pai] -= 2;
        return true;
    }
    bool pick_koutsu(const Hand& hand, Pai pai, Hand& nhand)
    {
        if(MAP_GET(hand, pai) < 3) return false;

        nhand = hand;
        nhand[pai] -= 3;
        return true;
    }
    bool pick_shuntsu(const Hand& hand, Pai pai, Hand& nhand)
    {
        if(!pai_is_shunstart(pai)) return false;
        Pai p1 = pai_next_seq(pai);
        Pai p2 = pai_next_seq(p1);
        if(!MAP_GET(hand, pai) || !MAP_GET(hand, p1) || !MAP_GET(hand, p2)) return false;

        nhand = hand;
        --nhand[pai];
        --nhand[p1];
        --nhand[p2];
        return true;
    }

    const PickFunc PICKERS_MENTSU[] = { pick_koutsu, pick_shuntsu };

    bool has_machi_pair(const Hand& hand)
    {
        for(Pai p = PAI_MAN1; p != PAI_INVALID; p = pai_next_seq(p)){
            if(!MAP_GET(hand, p)) continue;
            // シャボ
            if(MAP_GET(hand, p) >= 2) return true;
            // 嵌張、両面、辺張
            if(MAP_GET(hand, p) == 1){
                if(!pai_is_ryanpenstart(p)) continue;
                // 両面or辺張
                Pai p1 = pai_next_seq(p);
                if(MAP_GET(hand, p1)) return true;
                // 嵌張
                if(pai_is_shunstart(p)){
                    Pai p2 = pai_next_seq(p1);
                    if(MAP_GET(hand, p2)) return true;
                }
            }
        }
        return false;
    }

    Result mentsu3(const Hand& hand, bool hand14)
    {
        // まず頭を取る。取れなければノーテン
        for(Pai p = PAI_MAN1; p != PAI_INVALID; p = pai_next_seq(p)){
            Hand nhand;
            // 頭を取れて、かつ2牌の待ち形(シャボ、嵌張、両面、辺張)を作れた
            // らテンパイ
            if(pick_toitsu(hand, p, nhand))
                if(has_machi_pair(nhand)) return RESULT_TENPAI;
        }
        return RESULT_NONE;
    }

    Result mentsu4(const Hand& hand, bool hand14)
    {
        // 13牌なら単騎待ち確定
        if(!hand14) return RESULT_TENPAI;

        // 14牌なら頭があれば和了、なければテンパイ
        for(Pai p = PAI_MAN1; p != PAI_INVALID; p = pai_next_seq(p))
            if(MAP_GET(hand, p) == 2) return RESULT_AGARI;
        return RESULT_TENPAI;
    }

    Result hand_std_do(const Hand& hand, Pai pai, int n_mentsu, bool hand14, Result least)
    {
        // 面子を取る
        for(Pai p = pai; p != PAI_INVALID; p = pai_next_seq(p)){
            for(size_t i = 0; i < NELEMS(PICKERS_MENTSU); ++i){
                Hand nhand;
                PickFunc picker = PICKERS_MENTSU[i];
                // 結果が確定したらそこで探索打ち切り
                // 14牌なら和了、13牌ならテンパイで結果が確定する
                // 14牌のテンパイについてはまだ和了の可能性があるので、最低
                // でもテンパイであることを記録した上でさらに探索を進める
                if(picker(hand, p, nhand)){
                    Result result = hand_std_do(nhand, p, n_mentsu+1, hand14, least);
                    if(result == RESULT_AGARI || (!hand14 && result == RESULT_TENPAI))
                        return result;
                    if(hand14 && result == RESULT_TENPAI)
                        least = RESULT_TENPAI;
                }
            }
        }

        // ここまでで面子は取れるだけ取っている
        // 3面子未満ならテンパイにもならない
        if(n_mentsu < 3){
            return least;
        }
        // 3面子
        else if(n_mentsu == 3){
            Result result = mentsu3(hand, hand14);
            if(result == RESULT_NONE) return least;
            else return result;
        }
        // 4面子
        else if(n_mentsu == 4){
            return mentsu4(hand, hand14);
        }
        else{
            throw logic_error("n_mentsu > 4");
        }
    }

    Result hand_std(const Hand& hand, bool hand14)
    {
        return hand_std_do(hand, PAI_MAN1, 0, hand14, RESULT_NONE);
    }

    Result hand_toi7(const Hand& hand)
    {
        int n = 0;
        for(Pai p = PAI_MAN1; p != PAI_INVALID; p = pai_next_seq(p)){
            // 同種牌が4枚ある場合、絶対に七対子テンパイにはならない
            // このチェックで七対子4枚使い形を弾く
            if(MAP_GET(hand, p) == 4) return RESULT_NONE;
            if(MAP_GET(hand, p) >= 2) ++n;
        }
        if(n == 7) return RESULT_AGARI;
        else if(n == 6) return RESULT_TENPAI;
        else return RESULT_NONE;
    }

    Result hand_kokushi(const Hand& hand, bool hand14)
    {
        int kind = 0;
        int num = 0;
        for(Pai p = PAI_MAN1; p != PAI_INVALID; p = pai_next_seq(p)){
            if(pai_is_yao(p)){
                int n = MAP_GET(hand, p);
                if(n){
                    ++kind;
                    num += n;
                }
            }
        }
        if(kind == 13) return (hand14 ? RESULT_AGARI : RESULT_TENPAI);
        else if(kind == 12 && num >= 13) return RESULT_TENPAI;
        else return RESULT_NONE;
    }

    Result hand14_chk(const Hand& hand)
    {
        Result result;
        if((result = hand_kokushi(hand, true)) != RESULT_NONE) return result;
        if((result = hand_toi7(hand)) != RESULT_NONE) return result;
        return hand_std(hand, true);
    }

    Result hand13_chk(const Hand& hand)
    {
        Result result;
        if((result = hand_kokushi(hand, false)) != RESULT_NONE) return result;
        if((result = hand_toi7(hand)) != RESULT_NONE) return result;
        return hand_std(hand, false);
    }
}

bool pai_is_valid(int pai)
{
    const Pai* last = PAI_LIST + NELEMS(PAI_LIST);
    return find(PAI_LIST, last, pai) != last;
}

bool pai_is_man(Pai pai) { return !(pai&0xF0); }
bool pai_is_pin(Pai pai) { return pai&0x10; }
bool pai_is_sou(Pai pai) { return pai&0x20; }
bool pai_is_kaze(Pai pai) { return pai&0x40; }
bool pai_is_sangen(Pai pai) { return pai&0x80; }

bool pai_is_ji(Pai pai) { return pai&0xC0; }
bool pai_is_suu(Pai pai) { return !pai_is_ji(pai); }

bool pai_is_yao(Pai pai)
{
    if(pai_is_ji(pai)) return true;
    int n = pai_get_number(pai);
    return (n == 0x01 || n == 0x09);
}
bool pai_is_tan(Pai pai) { return !pai_is_yao(pai); }

int pai_get_number(Pai pai)
{
    if(!pai_is_suu(pai)) throw invalid_argument("Not suu");
    return pai&0x0F;
}

bool pai_is_shunstart(Pai pai)
{
    return pai_is_suu(pai) && pai_get_number(pai) <= 7;
}

bool pai_is_ryanpenstart(Pai pai)
{
    return pai_is_suu(pai) && pai_get_number(pai) <= 8;
}

Pai pai_next_seq(Pai pai)
{
    if(pai == PAI_MAN9) return PAI_PIN1;
    else if(pai == PAI_PIN9) return PAI_SOU1;
    else if(pai == PAI_SOU9) return PAI_TON;
    else if(pai == PAI_PEI) return PAI_HAKU;
    else if(pai == PAI_CHUN) return PAI_INVALID;
    else return static_cast<Pai>(pai+1);
}

Result hand_chk(const Hand& hand)
{
    switch(hand_size(hand)){
    case 14: return hand14_chk(hand);
    case 13: return hand13_chk(hand);
    default: throw logic_error("Invalid hand size");
    }
}
